import os
import fnmatch
import librosa

def find_files(directory, regex):
    matching_files = []

    for root, _, files in os.walk(directory):
        for file in files:
            if fnmatch.fnmatch(file, regex):
                # Add the full path of the matching file to the list
                matching_files.append(os.path.join(root, file))

    return matching_files


def find_total_wer_files(directory):
    return find_files(directory, '*TOTAL_WER.txt')
    

def find_wav_files(directory):
    return find_files(directory, '*.wav')

def find_label_files(directory):
    return find_files(directory, '*-*.txt')



import librosa
import soundfile as sf  
import os
import shutil
import fnmatch
import json

def slice_wav_file(input_path, output_path, offset_time, start_time, end_time):
    # Load the WAV file
    audio, sample_rate = librosa.load(input_path, sr=None)

    # Convert the start and end times from seconds to samples
    offset = int(offset_time * sample_rate)
    start_sample = int(start_time * sample_rate)
    end_sample = int(end_time * sample_rate)
    length_sample = int(end_time * sample_rate)

    # Slice the audio
    sliced_audio = audio[start_sample-offset:end_sample-offset]

    # Save the sliced audio as a new WAV file
    sf.write(output_path, sliced_audio, sample_rate)


def parse_block(sound_path, label_path, trials_dir, verbose=False):
    if (verbose):
        print("PARSING BLOCK")

    #Parse the label file into a list of sentences and start times
    lines = []
    with open(label_path, 'r') as file:
        for line in file.readlines():
            lines.append(line.strip().split(', '))  # Use strip() to remove leading and trailing whitespace
    mike_onset = float(lines[1][1]) 
    mike_offset = float(lines[-1][1])
    sentences = []
    for line in lines[2:-2]:
        sentences.append([line[0], line[1].strip()])

    # Splitting the path and getting the title
    title = os.path.basename(label_path)[:-4].replace(".", "_")

    # Creating the directory paths
    trial_dir = os.path.join(trials_dir, title)
    trial_subdir = os.path.join(trial_dir, "trials")

    # Checking if the directory exists and removing it if it does
    if os.path.exists(trial_dir):
        shutil.rmtree(trial_dir)

    # Creating the directories
    os.makedirs(trial_dir)
    os.makedirs(trial_subdir)


    #Slice Sentences
    for i in range(len(sentences) - 1):
        save_sentence_path = os.path.join(trial_subdir, sentences[i][0].replace(".", "-"))

        start = float(sentences[i][0])
        length = float(sentences[i+1][0]) - start
        slice_wav_file(sound_path, save_sentence_path + ".wav", mike_onset, start, start + length)
        with open(save_sentence_path + ".txt", "w") as f:
            f.write(f"{sentences[i][1]} {sentences[i][1]}")

    #for the last sentence, slice it using the offset
    save_sentence_path = os.path.join(trial_subdir, sentences[-1][0].replace(".", "-"))
    start = float(sentences[-1][0])
    length = mike_offset - start
    slice_wav_file(sound_path, save_sentence_path + ".wav", mike_onset, start, start + length)
    with open(save_sentence_path + ".txt", "w") as f:
        f.write(f"{sentences[-1][1]} {sentences[-1][1]}")

    #write the transcription method json
    transcription_method_path = os.path.join(trial_dir, "transcription_method.json")
    transcription_method_dict = {}
    for sentence in sentences:
        transcription_method_dict[sentence[0]] = "original"
    with open(transcription_method_path, "w") as f:
        json.dump(transcription_method_dict, f, indent=4)

    if (verbose):
        print("PARSING COMPLETE")

    return trial_dir



# def duplicate_sentence_strings(label_path):
#     #parse the file for information
#     lines = []
#     with open(label_path, 'r') as file:
#         for line in file:
#             lines.append(line.strip().split(', '))  # Use strip() to remove leading and trailing whitespace
        
#     sentences = [(line[0], line[1]) for line in lines[2:-2]]
#     with open(label_path, 'w') as file:
#         file.write("mike-on times\n")
#         file.write(lines[1][0] + ", " + lines[1][1] + "\n")

#         for sentence in sentences:
#             file.write(f"{sentence[0]}, {sentence[1]}, {sentence[1]}\n")

#         file.write("mike-off times\n")
#         file.write(lines[-1][0] + ", " + lines[-1][1] + "\n")

# def find_txt_files(directory):
#     wav_files = []

#     for root, _, files in os.walk(directory):
#         for file in files:
#             if fnmatch.fnmatch(file, '*.txt'):
#                 # Add the full path of the matching file to the list
#                 wav_files.append(os.path.join(root, file))

#     return wav_files

# label_files = find_txt_files(r"C:\makin_temp_repo\align_phonemes\data\human_data\speech")
# for file in label_files:
#     duplicate_sentence_strings(file)

#################### SLICE UP ONE BLOCK #########################################################
# path = r"C:\makin_temp_repo\align_phonemes\data\human_data\speech\speech_results_Aug_19_16.52.44"
# audio_path = path + ".wav"
# label_path = path + ".txt"
# slice_human_data(audio_path, label_path)


#################### GET ONE SLICE OF AN AUDIO FILE #########################################################
# offset = 17.011275900003966 #(16.998701000004075 + 17.011275900003966) / 2.0
# start = 71.156168800021987
# length = 3.0
# source_path = r"C:\makin_temp_repo\align_phonemes\data\human_data\speech\speech_results_Aug_18_15.45.59.wav"
# save_path = r"C:\makin_temp_repo\align_phonemes\data\human_data\cut_speech\cut2.wav"

# slice_wav_file(source_path, save_path, offset, start, length)


import docker
import sys
import subprocess
import os

""" def stream_container_logs(container):
    # Stream and print the container's stdout in real-time
    for log_chunk in container.logs(stream=True, stdout=True, stderr=False):
        print(log_chunk.decode(), end='') """

def download_models(container):
    try:
        # Run MFA commands inside the Docker container
        mfa_command = [
            "mfa",
            "model",
            "download",
            "acoustic",
            "english_us_arpa"
        ]
        exec_result = container.exec_run(mfa_command)
        print("Command output:", exec_result.output.decode())

        mfa_command = [
            "mfa",
            "model",
            "download",
            "dictionary",
            "english_us_arpa"
        ]
        exec_result = container.exec_run(mfa_command)
        print("Command output:", exec_result.output.decode())

    except subprocess.CalledProcessError as e:
        print(f"Error executing MFA command: {e}")
        # Handle the error as needed

def validate_data(container, container_data_directory):
    try:
        # Run MFA commands inside the Docker container
        mfa_command = [
            "mfa",
            "validate",
            container_data_directory,
            "english_us_arpa",
            "english_us_arpa"
        ]
        exec_result = container.exec_run(mfa_command)
        print("Command output:", exec_result.output.decode())

    except subprocess.CalledProcessError as e:
        print(f"Error executing MFA command: {e}")
        # Handle the error as needed

def align_data(container, container_data_directory, container_aligned_data_directory):
    try:
        # Run MFA commands inside the Docker container
        mfa_command = [
            "mfa",
            "align",
            container_data_directory,
            "english_us_arpa",
            "english_us_arpa",
            container_aligned_data_directory
        ]
        exec_result = container.exec_run(mfa_command)
        print("Command output:", exec_result.output.decode())

    except subprocess.CalledProcessError as e:
        print(f"Error executing MFA command: {e}")
        # Handle the error as needed

def run_mfa(trial_directory, verbose=False):
    if verbose:
        print("RUNNING MFA")

    local_data_directory = os.path.join(trial_directory, "trials")
    local_aligned_data_directory = os.path.join(trial_directory, "textgrids")

    os.mkdir(local_aligned_data_directory)

    # Create a Docker client
    client = docker.from_env()

    # Specify the image name and tag
    image_name = "mmcauliffe/montreal-forced-aligner"
    image_tag = "latest"

    # Pull the Docker image
    client.images.pull(image_name, tag=image_tag)

    container_data_directory = "/data"
    container_aligned_data_directory = "/aligned"
    
    volume_mapping = {
        local_data_directory: {'bind': container_data_directory, 'mode': 'rw'},
        local_aligned_data_directory: {'bind': container_aligned_data_directory, 'mode': 'rw'}
    }

    # Run the Docker container
    container = client.containers.run(
        f"{image_name}:{image_tag}",
        command="/bin/bash",  # You can customize the command if needed
        volumes=volume_mapping,
        stdin_open=True,
        tty=True,
        detach=True
    )

    download_models(container)

    validate_data(container, container_data_directory)

    align_data(container, container_data_directory, container_aligned_data_directory)

    # Stop and remove the Docker container
    container.stop()
    container.remove()

    if (verbose):
        print("MFA COMPLETE")

    return local_aligned_data_directory



import src.funcs as f
import pandas as pd
import numpy as np
from pydub import AudioSegment
from jiwer import cer

class Evaluate_Model():
    def __init__(self, wav, txt):
        self.txt_expected = txt
        self.audio = AudioSegment.from_wav(wav)
        self.char_error_rate = []

    def splice_audio(self):
        self.spliced_audio = []
        times = np.array(self.df.Time)

        for i, s in enumerate(times):
            start = s * 1000    #millisecond
            end = start + self.offset * 1000
           # print(start, end)
            audio_seg = self.audio[start:end]
            #audio_seg.export(f'spliced_audio/chunk{i}', format='wav')
            self.spliced_audio.append(audio_seg)
        #print(self.spliced_audio)

    def parse_txt(self):
        self.df = pd.read_csv(self.txt_expected, sep=',', header=None, names=['Time', 'TXT'], skiprows=1)
    
        try:
            self.df = self.df.iloc[range(0, self.df[self.df['Time'] == 'mike-off times'].index[0])]
            self.df['Time'] = pd.to_numeric(self.df['Time'])
            on = self.df.iloc[0]
            #print(on)
            self.df.drop(index=self.df.index[0], axis=0, inplace=True)
            self.df['Time'] = self.df['Time'] - (on['Time'] + pd.to_numeric(on['TXT']))/2
            #print(self.df)
        except:
            print("TXT files contain unexpected formatting")
        trials_time = self.df['Time']
        trials_time = trials_time.diff()
        metrics = {'Mean': trials_time.mean(), 'Std': trials_time.std(), 'Var': trials_time.var()}
        self.trial_length = metrics['Mean']
        self.offset = round(metrics['Mean'] / 3)
       # print(self.offset)
        if metrics['Var'] >= 0.1:
            print("Warning: variance of the trials length is high (>= 0.1)")

    def evaluate_transcript(self):
        file = open('transcripts', 'w')
        transcripts = []
        for index, chunk in enumerate(self.spliced_audio):
            #print(chunk)
            samples = chunk.get_array_of_samples()
            #print(samples)
            samples = np.array(samples).astype(np.float32)/32768 # 16 bit
            #print(self.df['TXT'].loc[index])
            try:
                transcript = f.get_transcription(samples, sr=chunk.frame_rate)
            except Exception as err:
                print(f"Could not proceed running the model in {index} due to: ", err)
                break
            else:
                self.char_error_rate.append( cer(self.df['TXT'].loc[index + 1], transcript))
                # print(transcript)
                transcripts.append(transcript + '\n')
            
        self.avg_error = sum(self.char_error_rate)/len(self.char_error_rate)
        print("Character Error Rate", self.avg_error)
        file.writelines(transcripts)
        file.close()
        
    def evaluate_phonemes(self):
        file = open('phoneme_alignment.txt', 'w')
        alignment = []
        for index, chunk in enumerate(self.spliced_audio):
            samples = chunk.get_array_of_samples()
            samples = np.array(samples).astype(np.float32)/32768 # 16 bit
            try:
                phonemes = f.get_phoneme_alignments(samples, sr=chunk.frame_rate)
            except Exception as err:
                print(f"Could not run model for test {index} due to {err}")
            else:
                alignment.append(phonemes)
        file.writelines(alignment)
        file.close()



import os
import json
import jiwer
import string


def find_textgrid_files(directory):
    textgrid_files = []

    for root, dirs, files in os.walk(directory):
        for file in files:
            if file.endswith(".TextGrid"):
                textgrid_files.append(os.path.join(root, file))

    return textgrid_files


def extract_mike_times(file_path):
    with open(file_path, 'r') as file:
        print(file_path)
        lines = file.readlines()

        mike_on_index = lines.index("mike-on times\n") + 1
        mike_on_times = [float(num) for num in lines[mike_on_index].split(',')]

        mike_off_index = lines.index("mike-off times\n") + 1
        mike_off_times = [float(num) for num in lines[mike_off_index].split(',')]

    return mike_on_times, mike_off_times


def initialize_json(block_label_path, json_path):
    mike_on_times, mike_off_times = extract_mike_times(block_label_path)

    alignment_data = {
        "mike-on times": [],
        "mike-off times": [],
        "alignments": []
    }

    alignment_data["mike-on times"].extend(mike_on_times)
    alignment_data["mike-off times"].extend(mike_off_times)

    with open(json_path, 'w') as json_file:
        json.dump(alignment_data, json_file, indent=4)


def extract_phones_from_textgrid(textgrid_path):
    try:
        with open(textgrid_path, 'r') as file:
            lines = file.readlines()

        phones_tier_index = lines.index('    item [2]:\n')

        phone_intervals = []
        for line in lines[phones_tier_index + 5:]:
            if line.startswith('            xmin'):
                xmin = float(line.split('=')[1].strip())
            elif line.startswith('            xmax'):
                xmax = float(line.split('=')[1].strip())
            elif line.startswith('            text'):
                text = line.split('=')[1].strip().strip('"')
                phone_intervals.append((xmin, xmax, text))

        return phone_intervals

    except Exception as e:
        print(f"Error: {e}")
        return None


def clean_phones(phone_intervals):
    cleaned_phones = []

    for xmin, xmax, text in phone_intervals:
        cleaned_text = text.rstrip('012')
        cleaned_text = "[SIL]" if cleaned_text == "" else cleaned_text
        cleaned_phones.append((xmin, xmax, cleaned_text))

    return cleaned_phones


def get_transcript_word_list(transcript_file):
    try:
        with open(transcript_file, 'r', encoding='utf-8') as file:
            sentence = file.readline().strip()

            translator = str.maketrans("", "", string.punctuation)
            processed_sentence = sentence.translate(translator).lower()

            word_list = processed_sentence.split()

            return word_list

    except Exception as e:
        print(f"Error: {e}")

        return None


def find_split_index(input_list):
    lowest_wer = 1
    best_index = 0

    for index in range(len(input_list)):

        first_half = ' '.join(input_list[:index])
        second_half = ' '.join(input_list[index:])

        if first_half and second_half:
            test_wer = jiwer.wer(first_half, second_half)

            if test_wer < lowest_wer:
                lowest_wer = test_wer
                best_index = index

    return best_index, lowest_wer


def append_trial_to_json(json_path, base_filename, cleaned_phone_interval_list, phones_split_index,
                         transcript_word_list, words_split_index, sentence_method):
    first_sentence_words = ' '.join(transcript_word_list[:words_split_index])
    second_sentence_words = ' '.join(transcript_word_list[words_split_index:])

    first_sentence_phonemes = cleaned_phone_interval_list[:phones_split_index]
    second_sentence_phonemes = cleaned_phone_interval_list[phones_split_index:]

    try:
        with open(json_path, 'r') as json_file:
            data = json.load(json_file)

        new_alignment = {
            base_filename: {
                "used sentence method": sentence_method,
                "used-sentence 1": first_sentence_words,
                "phoneme list 1": first_sentence_phonemes,
                "used-sentence 2": second_sentence_words,
                "phoneme list 2": second_sentence_phonemes
            }
        }

        data["alignments"].append(new_alignment)

        with open(json_path, 'w') as json_file:
            json.dump(data, json_file, indent=4)

        return True

    except Exception as e:
        print(f"Error appending alignment: {e}")
        return False


def get_sentence_method():
    pass


def demarcate_to_json(trial_directory, block_label_path, textgrid_directory, json_path):
    transcript_directory = os.path.join(trial_directory, "trials")
    textgrid_directory = os.path.join(trial_directory, "textgrids")
    method_json = os.path.join(trial_directory, "transcription_method.json")
    methods_by_filename = {}
    with open(method_json, "r") as f:
        methods_by_filename = json.load(f)

    initialize_json(block_label_path, json_path)

    textgrid_file_list = find_textgrid_files(textgrid_directory)

    for textgrid_file in textgrid_file_list:
        base_filename = os.path.splitext(os.path.basename(textgrid_file))[0]
        transcript_file = os.path.join(transcript_directory, base_filename + ".txt")

        if not os.path.exists(transcript_file):
            print(textgrid_file + " contains no corresponding .txt file in " + transcript_directory)

        else:
            cleaned_phone_interval_list = clean_phones(extract_phones_from_textgrid(textgrid_file))
            transcript_word_list = get_transcript_word_list(transcript_file)

            phone_list = [phone_interval[2] for phone_interval in cleaned_phone_interval_list]

            phones_split_index, phones_split_wer = find_split_index(phone_list)
            words_split_index, words_split_wer = find_split_index(transcript_word_list)

            sentence_method = methods_by_filename[base_filename.replace("-", ".")]

            append_trial_to_json(json_path, base_filename, cleaned_phone_interval_list, phones_split_index,
                                 transcript_word_list, words_split_index, sentence_method)
    
    return json_path


import json
from scipy.io import wavfile
import numpy as np
from jiwer import wer
import os

from utils.find_files import find_wav_files, find_label_files
from src.wav2vec import get_transcription

def apply_wav2vec_transcriptions(trial_directory, verbose=False):
    if (verbose):
        print("APPLYING TRANSCRIPTIONS")

    label_paths = sorted(find_label_files(trial_directory))
    wav_paths = sorted(find_wav_files(trial_directory))
    transcription_methods = {}
    with open(os.path.join(trial_directory, "transcription_method.json"), "r") as f:
        json_string = f.read()
        transcription_methods = json.loads(json_string)

    for wav_path, label_path in zip(wav_paths, label_paths):
        original_transcript = ""
        with open(label_path, "r") as f:
            original_transcript = f.readline()

        samplerate, data = wavfile.read(wav_path)
        data = data.astype(np.float32)
        wav2vec_transcript = get_transcription(data, sr=samplerate).lower()

        start_time = os.path.basename(label_path)[:-4].replace("-", ".")
        transcription_methods[start_time] = "wav2vec"

        with open(label_path, "w") as f:
            f.write(wav2vec_transcript)
        
    with open(os.path.join(trial_directory, "transcription_method.json"), "w") as f:
        json.dump(transcription_methods, f, indent=4)
    
    if (verbose):
        print("TRANSCRIPTIONS APPLIED")


def apply_critical_error_transcriptions(trial_directory, critical_error, verbose=False):
    if (verbose):
        print("APPLYING TRANSCRIPTIONS")

    label_paths = sorted(find_label_files(trial_directory))
    wav_paths = sorted(find_wav_files(trial_directory))
    transcription_methods = {}
    
    with open(os.path.join(trial_directory, "transcription_method.json"), "r") as f:
        json_string = f.read()
        transcription_methods = json.loads(json_string)

    for wav_path, label_path in zip(wav_paths, label_paths):
        original_transcript = ""
        with open(label_path, "r") as f:
            original_transcript = f.readline()

        samplerate, data = wavfile.read(wav_path)
        data = data.astype(np.float32)
        wav2vec_transcript = get_transcription(data, sr=samplerate).lower()

        if wer(wav2vec_transcript, original_transcript) > critical_error:
            start_time = os.path.basename(label_path)[:-4].replace("-", ".")
            print(label_path)
            print(start_time)
            print(transcription_methods[start_time])
            transcription_methods[start_time] = "wav2vec"

            with open(label_path, "w") as f:
                f.write(wav2vec_transcript)
        
    with open(os.path.join(trial_directory, "transcription_method.json"), "w") as f:
        json.dump(transcription_methods, f, indent=4)
    
    if (verbose):
        print("TRANSCRIPTIONS APPLIED")
        


        


from setuptools import setup, find_packages

setup(
    name='align_phonemes',
    version='0.1',
    packages=find_packages(),
    description='A short description of your project',
    long_description=open('README.md').read(),
    author='Will Stonebridge, Tyler Dierckman, Alex De Bruyn',
    url='https://github.itap.purdue.edu/kumar603/align_phonemes',
    install_requires=[
        'librosa>=0.10.1',
        'soundfile>=0.12.1',
        'docker>=7.0.0',
        'jiwer>=3.0.3',
        'scipy>=1.11.3',
        'numpy>=1.26.1',
        'torch>=2.1.0',
        'transformers>=4.34.1'
    ],
    classifiers=[
        # Classifiers help users find your project by categorizing it
        # For a full list, see https://pypi.org/classifiers/
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    package_data={
        'align_phonemes': ['get_forced_alignments.py',
                           'utils/*', 'trials/*', 'src/*']
    },
    exclude_package_data={
        'align_phonemes': ['data/*', 'env/*', 'pretrained_models/*'],
    }
)


import librosa
import numpy as np
import torch
from transformers import Wav2Vec2Processor, Wav2Vec2ForCTC

def get_transcription(mic_data, sr=16000):
    
    target_sr = 16000
    if sr != target_sr:
        mic_data = librosa.resample(np.float32(mic_data), orig_sr=sr, target_sr = 16000)
        
    model_name = "facebook/wav2vec2-large-960h-lv60-self"
    processor = Wav2Vec2Processor.from_pretrained(model_name)
    model = Wav2Vec2ForCTC.from_pretrained(model_name)
    
    input_values = processor(mic_data, sampling_rate = target_sr,
                             return_tensors="pt").input_values.to(torch.float32)
    
    # Normalize: Gave better accuracy
    input_values = (input_values - input_values.mean())/input_values.std()
    
    with torch.no_grad():
        logits = model(input_values).logits
        
    predicted_ids = torch.argmax(logits, dim=-1)
    transcription = processor.batch_decode(predicted_ids)[0]
    
    return transcription